/*
	Copyright  1995-2016, The AROS Development Team. All rights reserved.
	$Id$
*/

/**********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <intuition/intuition.h>

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dostags.h>
#include <graphics/gfxbase.h>
#include <graphics/rpattr.h>
#include <cybergraphx/cybergraphics.h>
#include <intuition/imageclass.h>
#include <intuition/icclass.h>
#include <intuition/gadgetclass.h>
#include <intuition/cghooks.h>
#include <datatypes/datatypesclass.h>
#include <datatypes/pictureclass.h>

#include <clib/alib_protos.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/utility.h>
#include <proto/iffparse.h>
#include <proto/datatypes.h>

#include <aros/symbolsets.h>

#include "debug.h"
#include "methods.h"

//#include <tiff.h>
#include <tiffio.h>

/* Open superclass */
ADD2LIBS("datatypes/picture.datatype", 0, struct Library *, PictureBase);

/**********************************************************************/

static int ReadNextInt(BPTR FileHandle)
{
	int c;
	int retval = 0;

	c = FGetC(FileHandle);
	while (c==' ' || c=='\t' || c=='\r' || c=='\n' || c==12) 
	{
		c = FGetC(FileHandle);
	}
	
	retval = c - '0';
	c = FGetC(FileHandle);
	while (c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=12 && c!=EOF) 
	{
		retval = retval * 10 + (c - '0');
		c = FGetC(FileHandle);
	}
	
	return retval;	
}

/**************************************************************************************************/

void ConvertRGBAImage2RGB(unsigned char *RGBAImage, unsigned char *RGBTemp, LONG wd, LONG ht)
{    
	//Convert RGBA to RGB to Save As 24bit.
        
	int x, y;    
	int source = 0;
	int target = 0;
	unsigned char a,r,g,b; 
	
	for (y = 0; y < ht; y++)
	{ 
		for (x = 0; x < wd; x++)
		{    
			r = (RGBAImage[source]);          
			g = (RGBAImage[source+1]);  
			b = (RGBAImage[source+2]); 
			a = (RGBAImage[source+3]); 
			
			//Convert RGBA to RGB
			RGBTemp[target] = r;
			RGBTemp[target+1] = g;
			RGBTemp[target+2] = b;
			//RGBTemp[target+3] = a;			
						
			target += 3;
			source += 4;
		}
	}
}

/**************************************************************************************************/
static BOOL Load_TIFF(struct IClass *cl, Object *o)
{
	BOOL err;
        ULONG row;
        ULONG w, h;
	ULONG* raster;
	IPTR sourcetype;
	BPTR FileHandle;	
	BOOL retval = FALSE;
	LONG pconfig, pmetric;
        STRPTR namebuffer = NULL;        
	struct BitMapHeader *bmhd;        
	

	D(bug("tiff.datatype/LoadTIFF(): Starting. (24 bit mode)\n"));	

	
	 if( GetDTAttrs(o,   DTA_SourceType    , (IPTR)&sourcetype ,
			DTA_Handle        , (IPTR)&FileHandle,	
			PDTA_BitMapHeader , (IPTR)&bmhd,
			TAG_DONE) != 3 )
	{
		D(bug("tiff.datatype/LoadTIFF(): GetDTAttrs(DTA_Handle, DTA_BitMapHeader) error !\n"));
		SetIoErr(ERROR_OBJECT_NOT_FOUND);
		return FALSE;
	}
	D(bug("tiff.datatype/LoadTIFF(): GetDTAttrs(DTA_Handle, DTA_BitMapHeader) successful\n"));
	
	if ( sourcetype == DTST_RAM && FileHandle == BNULL )
	{
		D(bug("tiff.datatype/LoadTIFF(): Creating an empty object\n"));
		return TRUE;
	}
	if ( sourcetype != DTST_FILE || !FileHandle || !bmhd )
	{
		D(bug("tiff.datatype/LoadTIFF(): Unsupported sourcetype mode\n"));
		SetIoErr(ERROR_NOT_IMPLEMENTED);
		return FALSE;
	}
	
	
	// Check file Signature. Read file header
	//Seek(FileHandle, 0, OFFSET_BEGINNING);
	D(bug("tiff.datatype/LoadTIFF(): Seek was successful\n"));
    
	//Read TIFF Header: 'I I' or 'M M' (Intel or Motorola Byte Order)
	//Read TIFF Version: 42
	//Read TIFF FirstOffset: (LONG Offset to First Image IFD Entry)
	
	
	//* Setup raster image buffer for ULONG RGBA pixel array *//
	//raster=AllocVec(Width*3, MEMF_ANY | MEMF_CLEAR);
	LONG npixels= w*h;
	raster=(ULONG *) _TIFFmalloc(npixels *sizeof(ULONG));	
	if(!raster)
	{
		D(bug("tiff.datatype/LoadTIFF(): AllocVec(RGBA Buffer) failed.\n"));
		SetIoErr(ERROR_NO_FREE_STORE);
		return FALSE;
	}
	D(bug("tiff.datatype/LoadTIFF(): RGBA buffer successfully allocated.\n"));
	
	//* Setup RGB Image Buffer for RGBA pixel array *//	
	unsigned char *RGBImage = NULL;
	
        //RGBImage = (unsigned char *)AllocVec(npixels * 3,MEMF_ANY);  
	//RGBImage=(unsigned char *) _TIFFmalloc(npixels *3);
	//if(!RGBImage)
	//{
		//D(bug("tiff.datatype/LoadTIFF(): AllocVec(RGB Image Buffer) failed.\n"));
		//SetIoErr(ERROR_NO_FREE_STORE);
		//return FALSE;
	//}
	//D(bug("tiff.datatype/LoadTIFF(): RGB Image Buffer successfully allocated.\n"));
	
    
	//**  Convert filehandle **//
	D(bug("\ntiff.datatype/LoadTIFF(): Attempting to Get File Name from FileHandle.\n"));  
	
	// NameFromLock(BPTR lock, STRPTR buffer, LONG length )
	namebuffer = malloc(256 * sizeof(char));
	err= NameFromFH(FileHandle, namebuffer, 256);
	if (err != 0)
	{
		D(bug("\ntiff.datatype/LoadTIFF(): File Name: %s\n\n", namebuffer));                  
	}
	else
	{
		D(bug("\ntiff.datatype/LoadTIFF(): Failed to get File Name.\n"));
		SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		return FALSE;
	}
	
    
	// Open Tiff file to get new filehandle
	TIFF* tif = TIFFOpen(namebuffer,"r");
	if (tif)
	{
		D(bug("tiff.datatype/LoadTIFF(): Successfully Opened Tiff File.\n"));
        
		//#define uint32 unsigned long - ULONG
		TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
		TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
		//TIFFGetField(tif, TIFFTAG_PLANARCONFIG, &pconfig);
		TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &pmetric);
  	    
		//D(bug("\ntiff.datatype/LoadTIFF(): Picture Data: Width %d x Height %d\n", w, h);

		//Photometric must be RGB.
		
		//TIFFReadRGBAImageOriented( ... ); //Use raster RGBA Image Buffer
		 TIFFReadRGBAImageOriented(tif, w, h, raster, ORIENTATION_TOPLEFT, 0);
		
		//Condense Rows //Remove Alpha //Use (unsigned char *)raster & RGB Image Buffer
		//void ConvertRGBAImage2RGB( unsigned char *RGBAImage, unsigned char *RGBTemp, LONG wd, LONG ht )
		ConvertRGBAImage2RGB( (unsigned char *)raster, RGBImage, w, h );
		 
		//Close TIFF
		TIFFClose(tif);
		//retval = TRUE;
	}
	else
	{
		D(bug("tiff.datatype/LoadTIFF(): Not a valid TIFF File\n"));
		SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		return FALSE;
	}
	//Free raster image buffer
	_TIFFfree(raster); 
	//Free namebuffer
	
    
        //* Set Width & Height *//
	bmhd->bmh_Width  = w;
	bmhd->bmh_Height = h;
	bmhd->bmh_PageWidth = bmhd->bmh_Width;
	bmhd->bmh_PageHeight = bmhd->bmh_Height;

	D(bug("tiff.datatype/LoadTIFF(): Using 24 bit colors\n"));
	bmhd->bmh_Depth = 24;
	
	

	//* Setup RGB Row Buffer for scanline pixel array *//
	unsigned char *RGBBuffer = NULL;  
	//RGBBuffer=AllocVec(w*3, MEMF_ANY | MEMF_CLEAR);
	RGBBuffer=(unsigned char *) _TIFFmalloc(w *3);
	if(!RGBBuffer)
	{
		D(bug("tiff.datatype/LoadTIFF(): AllocVec(RGBBuffer) failed\n"));
		SetIoErr(ERROR_NO_FREE_STORE);
		return FALSE;
	}
	D(bug("tiff.datatype/LoadTIFF(): RGB buffer successfully allocated\n"));
	
	
	
	//* Write RGB image buffer to new datatype object. PBPAFMT_RGB *//
	D(bug("\ntiff.datatype/OM_NEW: WRITEPIXELARRAY of whole picture\n"));	
	
	LONG source_offset;		
	LONG rgbSize = (w *3);
	for(row=0; row<h; row++)
	{
		//Copy RGB Rows from RGB Image Buffer to RGB Row Buffer
		source_offset = (row * rgbSize);		
		memcpy(RGBBuffer, RGBImage + source_offset, rgbSize);
			
		//Write RGB Row to new datatype object.
		if(!DoSuperMethod(cl, (Object *) o,
						PDTM_WRITEPIXELARRAY,	/* Method_ID */
						(IPTR) RGBBuffer,	/* PixelData */
						PBPAFMT_RGB,		/* PixelFormat */
						w*3,		/* PixelArrayMod (number of bytes per row) */
						0,			/* Left edge */
						row,			/* Top edge */
						w,			/* Width */
						1))			/* Height (here: one line) */
		{
			D(bug("tiff.datatype/OM_NEW: WRITEPIXELARRAY failed\n"));
			FreeVec(RGBBuffer);
			return FALSE;
		}
	}
	D(bug("\ntiff.datatype/OM_NEW: WRITEPIXELARRAY of whole picture done\n"));
		
		
	_TIFFfree(RGBBuffer); 	
	//FreeVec(RGBBuffer);
	//FreeVec(RGBImage);

	//SetDTAttrs((Object *) o, NULL, NULL, DTA_ObjName,      (IPTR) Title,
	//DTA_NominalHoriz, Width,
	//DTA_NominalVert,  Height,
	//TAG_DONE);

	D(bug("tiff.datatype/OM_NEW: Ending. (24 bit mode)\n"));
	return TRUE;
}

/**************************************************************************************************/

static BOOL SavePNM(struct IClass *cl, Object *o, struct dtWrite *dtw )
{
	BPTR		    filehandle;
	unsigned int    width, height, numplanes, y;
	UBYTE		    *linebuf;
	struct BitMapHeader     *bmhd;
	long                    *colorregs;

	D(bug("ppm.datatype/SavePNM()\n"));

	/* A NULL file handle is a NOP */
	if( !dtw->dtw_FileHandle )
	{
		D(bug("ppm.datatype/SavePNM() --- empty Filehandle - just testing\n"));
		return TRUE;
	}
	filehandle = dtw->dtw_FileHandle;

	/* Get BitMapHeader and color palette */
	if( GetDTAttrs( o,  PDTA_BitMapHeader, (IPTR) &bmhd,
				PDTA_CRegs,        (IPTR) &colorregs,
				TAG_DONE ) != 2UL ||
			!bmhd || !colorregs )
	{
		D(bug("ppm.datatype/SavePNM() --- missing attributes\n"));
		SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		return FALSE;
	}

	width = bmhd->bmh_Width;
	height = bmhd->bmh_Height;
	numplanes = bmhd->bmh_Depth;
	if( numplanes != 24 )
	{
		D(bug("ppm.datatype/SavePNM() --- color depth %d, can save only depths of 24\n", numplanes));
		SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		return FALSE;
	}
	D(bug("ppm.datatype/SavePNM() --- Picture size %d x %d (x %d bit)\n", width, height, numplanes));

	/* Write header to file */
	if( FPrintf( filehandle, "P6\n#Created by AROS ppm.datatype aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa\n%ld %ld\n255\n",
				(long)width, (long)height ) == -1 )
	{
		D(bug("ppm.datatype/SavePNM() --- writing header failed\n"));
		return FALSE;
	}

	/* Now read the picture data line by line and write it to a chunky buffer */
	if( !(linebuf = AllocVec(width*3, MEMF_ANY)) )
	{
		SetIoErr(ERROR_NO_FREE_STORE);
		return FALSE;
	}
	D(bug("ppm.datatype/PPM_Save() --- copying picture with READPIXELARRAY\n"));
	for (y=0; y<height; y++)
	{
		if(!DoSuperMethod(cl, o,
					PDTM_READPIXELARRAY,	/* Method_ID */
					(IPTR)linebuf,		/* PixelData */
					PBPAFMT_RGB,		/* PixelFormat */
					width,			/* PixelArrayMod (number of bytes per row) */
					0,			/* Left edge */
					y,			/* Top edge */
					width,			/* Width */
					1))			/* Height */
		{
			D(bug("ppm.datatype/SavePNM() --- READPIXELARRAY line %d failed !\n", y));
			FreeVec(linebuf);
			SetIoErr(ERROR_OBJECT_WRONG_TYPE);
			return FALSE;
		}
		if( FWrite( filehandle, linebuf, width*3, 1 ) != 1 )
		{
			D(bug("ppm.datatype/SavePNM() --- writing picture data line %d failed !\n", y));
			FreeVec(linebuf);
			return FALSE;
		}
	}

	D(bug("ppm.datatype/SavePNM() --- Normal Exit\n"));
	FreeVec(linebuf);
	SetIoErr(0);
	return TRUE;
}

/**************************************************************************************************/

IPTR TIFF__OM_NEW(Class *cl, Object *o, Msg msg)
{
	Object *newobj;
	
	D(bug("pnm.datatype/DT_Dispatcher: Method OM_NEW\n"));
	
	newobj = (Object *)DoSuperMethodA(cl, o, msg);
	if (newobj)
	{
		if (!Load_TIFF(cl, newobj))
		{
			CoerceMethod(cl, newobj, OM_DISPOSE);
			newobj = NULL;
		}
	}

	return (IPTR)newobj;
}

/**************************************************************************************************/

IPTR TIFF__DTM_WRITE(Class *cl, Object *o, struct dtWrite *dtw)
{
	D(bug("pnm.datatype/DT_Dispatcher: Method DTM_WRITE\n"));
	if( (dtw -> dtw_Mode) == DTWM_RAW )
	{
		/* Local data format requested */
		return SavePNM(cl, o, dtw );
	}
	else
	{
		/* Pass msg to superclass (which writes an IFF ILBM picture)... */
		return DoSuperMethodA( cl, o, (Msg)dtw );
	}
}
